(     Title:  Forth Search-order tester
       File:  otester.fs
  Test File:  otester-test.fs
   Log File:  otester.log
     Author:  David N. Williams
    License:  Public Domain
    Version:  1.0.0
    Revised:  September 28, 2011

This ANS/ISO Forth library defines Hayes-style tests for the
search-order stack.  Words not declared under USER INTERFACE are
considered to be private, not guaranteed to be available in
future versions of otester.

USER INTERFACE

   tests:  O{  O->  }O
  errors:  OT-ERROR-XT  OT-ERROR-DEFAULT  OT-ERROR-INDEX
  arrays:  OT-INITIALS  OT-RESULTS  OT-GOALS
           /MAX-OT-A  DROPS
           OT-A!  OT-A@  OT-A-LAST@  OT-A<>
           RESTORE-OT-INITIALS

Basic usage takes the form:

  <oinitials> O{ <oresults> O-> <ogoals> }O <oinitials>

Here <oresults> and <ogoals> stand for any word sequences that
leave net results and goals wids, including none, on the
search-order stack, above any items present just before O{.

If <oresults> or <ogoals> corrupts the initial search order, an
exception is thrown.  Otherwise O{ saves the initial search
order, and O-> and }O restore it, removing, respectively, the
results and goals.  Thus O{ ... O-> ... }O leaves the initial
search order unchanged.

Results and goals are compared; and an error is reported by
executing the xt in the variable OT-ERROR-XT.  Either way, the
initial search order remains unchanged.

Before an error is reported, either the index into the result
and goal arrays at which the mismatch was found is stored in
OT-ERROR-INDEX, or zero is stored if the arrays have different
dimension.  OT-ERROR-INDEX is available to the user for more
sophisticated, error reporting.  An example of how to do that is
included in otester-test.fs.

All three words ignore any effects that <oresults> or <ogoals>
might have on other stacks.  The library can be combined with
ttester.fs to include tests of the data and floating-point
stacks, or with other compatible testers to include other
stacks.  See the files otester-test.fs and order-test.fs for
examples of how to use otester and ttester together.  Changes
from default error reporting must be revectored for both otester
and ttester.

This code does not change BASE.
)
BASE @ DECIMAL

s" WORDLISTS" ENVIRONMENT? 0= [IF]
  cr .( ***can't determine max search order size) ABORT [THEN]
( max#wordlists) cells CONSTANT /OT-MAX-ORDER


\ *** ERRORS

VARIABLE OT-ERROR-XT
: OT-ERROR OT-ERROR-XT @ EXECUTE ;  \ vectored error reporting

VARIABLE OT-ERROR-INDEX  \ holds array index|-1 after error

: OT-ERROR-DEFAULT  ( c-addr u -- )
(
Display an error message followed by the line that had the
error.
)
  type source type cr ;
' OT-ERROR-DEFAULT OT-ERROR-XT !


\ *** ARRAYS
(
The first array slot has index zero.
)

\ Arrays are headed by a cell-sized count.
CREATE OT-INITIALS  /OT-MAX-ORDER cell+ allot
CREATE OT-RESULTS   /OT-MAX-ORDER cell+ allot
CREATE OT-GOALS     /OT-MAX-ORDER cell+ allot

: OT-A!  ( wid_[n-1] ... wid_0 +n addr -- )
  over 1+ cells over + swap DO i ! 1 cells +LOOP ;

: OT-A@  ( addr -- wid_[n-1] ... wid_0 n )
  dup @ cells over + DO i @ -1 cells +LOOP ;

: OT-A-LAST@  ( 'a k -- wid_[n-1] ... wid_[n-k] k )
(
Let n be the dimension of the input array, and 0 <= k <= n.
Fetch the subarray of dimension k at its end.
)
   LOCALS| k 'a |
   k IF
     'a @ 1+ k - cells 'a + ( 'wid_[n-k])
     'a @ cells 'a + ( 'wid_[n-k] 'wid_[n-1])
     DO i @ -1 cells +LOOP
   THEN k ;

\ This definition would have saved a little space, but the one
\ above does not require a test for zero dimension.
\ : OT-A@  ( 'a -- wid_[n-1] ... wid_0 n )  dup @ OT-A-LAST@ ;

[UNDEFINED] DROPS [IF]
  : DROPS  ( xn ... x1 n -- )  0 ?DO drop LOOP ; [THEN]


true [IF]
: OT-A<>  ( wid_[n-1] ... wid_0 n addr -- [index+1]|-1|0 )
(
Nonzero return for errors:
  -1       unequal array size
  index+1  first unequal array elements, other testers return last
)
  2dup @ <> IF drop drops -1 EXIT THEN
  ( addr) dup @ 0
  LOCALS| index dim addr #left |
  BEGIN
    #left
  WHILE
    #left 1- to #left
    addr cell+ dup to addr
    @ =
  WHILE
  REPEAT
    dim #left - to index
  THEN #left drops index ;
[ELSE]
: OT-A<>  ( wid_[n-1] ... wid_0 n addr -- [index+1]|-1|0 )
(
Nonzero return for errors:
  -1       unequal array size
  index+1  deepest unequal array element pair
)
  2dup @ <> IF drop drops -1 EXIT THEN
  ( addr) dup @ 0
  LOCALS| index dim addr #left |
  dim cells addr + to addr
  BEGIN
    #left
  WHILE
    #left 1- to #left
    addr cell- dup to addr
    @ =
  WHILE
  REPEAT

**** FIX INDEX!!!
    dim #left - to index
  THEN #left drops index ;
[THEN]


\ *** TESTS

: RESTORE-OT-INITIALS  ( -- )  OT-INITIALS OT-A@ set-order ;

: O{  ( o: initials -- initials )
(
Store the initial search order and clear results and goals
search order storage, and the error case number.
)
  get-order OT-INITIALS OT-A!
  0 OT-RESULTS !   0 OT-GOALS ! ;

: O->  ( o: initials results -- initials )
(
Store the search order, including the possibly corrupted
initial order and any results above it, and restore the initial
search order.

Throw an exception if the total order has fewer wids than the
initial order, or if it has as many wids as the initials, but
the portion at its end does not agree with the initials.

Delete the inital order from the end of the stored total order,
leaving only the results.
)
  get-order restore-ot-initials
  OT-RESULTS OT-A!  \ save total order

  OT-INITIALS @ OT-RESULTS @ >
  ABORT" RESULTS UNDERFLOW START ORDER"

  OT-RESULTS OT-INITIALS @ OT-A-LAST@
  OT-INITIALS OT-A<>
  ABORT" RESULTS CLOBBER START ORDER"

  OT-INITIALS @ negate OT-RESULTS +! ;

: }O  ( o: initials goals -- initials )
(
Store the total search order, including the possibly corrupted
initial order and the goals above it, and restore the initial
search order.

Throw an exception if the total order has fewer wids than the
initial order, or if it has as many wids as the initials, but
the portion at its end does not agree with the initials.

Delete the initial order from the end of the stored total order,
leaving only the goals.

Compare the stored results and goals.  If their number is not
the same, set OT-ERROR-INDEX to 0, and exit through the OT-ERROR
vector.

Else if results and goals disagree, store the index at which
the first error is found in OT-ERROR-INDEX and exit through
OT-ERROR.
)
  get-order restore-ot-initials
  OT-GOALS OT-A!  \ save total order

  OT-INITIALS @ OT-GOALS @ >
  ABORT" GOALS UNDERFLOW START ORDER"

  OT-GOALS OT-INITIALS @ OT-A-LAST@ OT-INITIALS OT-A<>
  ABORT" GOALS CLOBBER START ORDER"

  OT-INITIALS @ negate OT-GOALS +!

  OT-GOALS @ OT-RESULTS @ <>
  IF -1 OT-ERROR-INDEX !
     s" WRONG NUMBER OF ORDER RESULTS: " OT-ERROR EXIT THEN 

  OT-RESULTS OT-A@ OT-GOALS OT-A<> ?dup
  IF 1- OT-ERROR-INDEX !
     s" INCORRECT ORDER RESULT: " OT-ERROR THEN ;

BASE !
\ END of otester.fs
